#ifndef RTCP_HEAP_H
#define RTCP_HEAP_H

#define _POSIX_C_SOURCE 200112L
#include "rtcproto/posixver.h"

#include <stdint.h>
#include <assert.h>
#include <errno.h>

#include "rtcproto/rtcproto.h"

/** User-defined comparison function for use in a generic heap.
*/
typedef int (*fn_heap_compare_t)(void*, void*);

/** Generic heap data structure.
  *
  * The comparison function is user-defined, so a "less" comparison will yield
  * a min-heap, and a "greater" comparison will yield a max-heap.
  *
  * A user-defined callback function will be invoked whenever a value is moved
  * into a new position on the heap. Currently called on every move.
  *
  * Expands its memory pool as required, but shrinks it lazily.
*/
typedef struct rtcp_heap_t
{
    /** Currently allocated capacity. */
    int capacity;
    /** Dynamically shrinking/expanding capacity. */
    int targetcapa;
    /** Currently used capacity. */
    int size;
    /** Size of the stored type, in bytes. */
    int typesize;
    /** Stored values. */
    void* values;
    /** Value swap storage. */
    void* swap;
    /** User-defined comparison function. */
    fn_heap_compare_t compare;
    /** User-defined value index update callback. */
    fn_idx_update_t callback;
    /** User pointer for the index update callback. */
    void* user;
} rtcp_heap_t;

void rtcp_heap_init(rtcp_heap_t* heap);
void rtcp_heap_cb(rtcp_heap_t* heap, fn_idx_update_t callback, void* user);
int rtcp_heap_malloc(rtcp_heap_t* heap, int capacity, int typesize, fn_heap_compare_t compare);
void rtcp_heap_free(rtcp_heap_t* heap);
int rtcp_heap_realloc(rtcp_heap_t* heap, int newcapa);
void rtcp_heap_swap(rtcp_heap_t* heap, int a, int b);

void rtcp_heap_tick(rtcp_heap_t* heap);
void rtcp_heap_use_size(rtcp_heap_t* heap);
int rtcp_heap_shrink(rtcp_heap_t* heap);

int rtcp_heap_property(rtcp_heap_t* heap);
void rtcp_heap_swim(rtcp_heap_t* heap, int idx);
void rtcp_heap_sink(rtcp_heap_t* heap, int idx);
int rtcp_heap_insert(rtcp_heap_t* heap, void* val);
void* rtcp_heap_peek(rtcp_heap_t* heap);
void rtcp_heap_extract(rtcp_heap_t* heap, void* val);
void* rtcp_heap_get(rtcp_heap_t* heap, int idx);

#endif /* RTCP_HEAP_H */
