#ifndef RTCP_HASH_H
#define RTCP_HASH_H

#define _POSIX_C_SOURCE 200112L
#include "rtcproto/posixver.h"

#include <stdint.h>
#include <assert.h>
#include <errno.h>

#include "rtcproto/rtcproto.h"
#include "rtcproto/romu.h"

/** User-defined hashmap iteration function.
*/
typedef void (*fn_hmap_iter_t)(void*);

/** Hashmap bucket states. A bucket can be empty, deleted, or occupied. */
enum {
    RTCP_HMAP_NIL = 0,
    RTCP_HMAP_DEL,
    RTCP_HMAP_OCC
};

/** Generic hashmap data structure.
  *
  * Uses open addressing, linear probing, and a load factor of 1/2.
  * Hashes keys with 3 iterations of the romu_mono32 pseudorandom generator.
  *
  * Expands its memory pool as required, but shrinks it lazily.
*/
typedef struct rtcp_hmap_t
{
    /** Currently allocated capacity. */
    int capacity;
    /** Dynamically shrinking/expanding capacity. */
    int targetcapa;
    /** Currently used capacity. */
    int size;
    /** Size of the stored type, in bytes. */
    int typesize;
    /** Bucket states. */
    char* records;
    /** Bucket keys. */
    fn_prthread_t* keys;
    /** Bucket values. */
    void* buckets;
} rtcp_hmap_t;

void rtcp_hmap_init(rtcp_hmap_t* hmap);
int rtcp_hmap_malloc(rtcp_hmap_t* hmap, int capacity, int typesize);
void rtcp_hmap_free(rtcp_hmap_t* hmap);
void rtcp_hmap_move(rtcp_hmap_t* dest, rtcp_hmap_t* src);
int rtcp_hmap_realloc(rtcp_hmap_t* hmap, int newcapa);

unsigned int rtcp_hmap_hashkey(fn_prthread_t key, int size);
float rtcp_hmap_loadfact(rtcp_hmap_t* hmap);

void rtcp_hmap_tick(rtcp_hmap_t* hmap);
void rtcp_hmap_use_size(rtcp_hmap_t* hmap);
int rtcp_hmap_shrink(rtcp_hmap_t* hmap);

int rtcp_hmap_add(rtcp_hmap_t* hmap, fn_prthread_t key, void* val);
void* rtcp_hmap_get(rtcp_hmap_t* hmap, fn_prthread_t key);
int rtcp_hmap_get_idx(rtcp_hmap_t* hmap, fn_prthread_t key);
void rtcp_hmap_del(rtcp_hmap_t* hmap, fn_prthread_t key);
void rtcp_hmap_iter(rtcp_hmap_t* hmap, fn_hmap_iter_t callback);

#endif /* RTCP_HASH_H */
