#ifndef RTCP_H
#define RTCP_H

/*
MIT License
Copyright © 2023 Benedict Blaise

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/** rtcproto: run-to-completion protothreads

PROTOTHREADS

Protothreads are extremely lightweight stackless threads designed for
severely memory constrained systems such as small embedded systems or
sensor network nodes. Protothreads can be used with or without an
underlying operating system.

Protothreads provide a blocking context on top of an event-driven
system, without the overhead of per-thread stacks. The purpose of
protothreads is to implement sequential flow of control without
complex state machines or full multi-threading.

The protothread implementation used here is the one originally written by
Adam Dunkels and Oliver Schmidt. I've lifted out only the parts relevant
to this project, and the changes were so slight that it still seemed sane to
include the license under which their protothread library was released under.

Their full license text can be found in "proto.h".

LOCAL STATES

One drawback of protothreads is the fact that their local continuation only
stores the point of execution in a function, and no local state whatsoever,
such as any automatic (stack) variables. The usual remedy for this in memory
constrained systems is to use static variables in functions, and to never have
more than one instance of a protothread running, which works well in contexts
such as the Contiki operating system's.

In the run-to-completion protothread scheduler, while it is still true that
only one protothread may be executing at any one time, this is only the result
of the style of scheduling it uses, its methods allow for more flexibility.
This means that the local states of each protothread invocation can be stored
separately while that particular invocation waits to be scheduled.

COOPERATIVE SCHEDULING

When the scheduler is expected to return control by a timestamp regardless of
whether the scheduled work is done, it provides protothreads with an extra
timestamp parameter by which they are also expected to return control.

This library aims to provide convenient macros that allow protothreads to
cooperate with their executing scheduler by mechanisms such as yielding
conditionally if a timepoint has passed, or generic loops that avoid the
overhead of clock checks on every iteration, checking only at every couple
iterations.

TICKING AND SHRINKING

Memory is shrunk lazily to avoid cyclical deallocation/allocation on
systems that require and relinquish capacity in regular bursts. This works
by ticking the data structure regularly, to let it know that a unit of time
has passed. Each time, the expected required capacity is decreased, and
when it falls below a threshold, the next write operation will reallocate
to a smaller footprint that accomodates the actual load.

PERFORMANCE STATISTICS

Important performance statistics are recorded for each protothread, to help
diagnose performance bottlenecks. See the documentation on the rtcp_stats_t
type for what these are. Before requesting these statistics, the protothread
function has to be registered with the scheduler either manually, or
automatically through queueing calls to that protothread.

SINGLE HEADER

The library is built with meson for the time being, but an up-to-date version
of a single-header rtcproto.h should be available in the source tree. If it
happens to be out of date, you can build it with "$ make".

All declarations should be at the top, and all definitions should be at the
bottom of this header. Do "#define RTCPROTO_IMPLEMENTATION" before including
this header in *exactly one* C file to create the implementation.

ANSI C & C99 & POSIX 2001

The language of the implementation itself is ANSI C (C89), but the library
depends on "stdint.h" for exact-width integer types (C99) and "unistd.h" for
measuring wall-clock time accurately with clock_gettime (SUSv2, POSIX.1-2001).
*/

#define _POSIX_C_SOURCE 200112L
#include "rtcproto/posixver.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <limits.h>
#include <unistd.h>

#include "rtcproto/proto.h"

/** Function executing a scheduled protothread.
  * All protothread functions must conform to this type if they are scheduled.
  \param proto protothread representation
  \param state user-defined local state associated with the protothread
  \param end NULL if the scheduler expects the protothread to complete,
             or a timestamp by which it is expected to suspend execution
*/
typedef char (*fn_prthread_t)(rtcp_pt_t* proto, void* state, struct timespec* end);

/** User-defined index update function.
  * Useful when two values are associated in two different data structures,
  * when they need to know the key/position of the other value in the
  * associated structure at all times.
  \param value stored element whose index was updated
  \param index the updated index of the element
  \param user optional user pointer
 */
typedef void (*fn_idx_update_t)(void* value, int index, void* user);

void rtcp_assert_always(int cond, char* fmt, ...);

unsigned int rtcp_ceil_power_of_two(unsigned int x);

int rtcp_capacity_from_size_and_factor(int size, float factor);

void rtcp_timespec_norm(struct timespec* spec);

int rtcp_timespec_diff_nsec(struct timespec* start, struct timespec* end);

int rtcp_timespec_less(struct timespec* start, struct timespec* end);

void rtcp_timespec_next(struct timespec* start);

void rtcp_timespec_prev(struct timespec* start);

struct timespec rtcp_timespec_add(struct timespec* start, struct timespec* end);

struct timespec rtcp_timespec_diff(struct timespec* start, struct timespec* end);

/* Clock-based protothread yielding for cooperative scheduling. */

/** Yield from the current protothread if a timepoint has passed.
  * Essentially, it acts as a conditional blocking wait to be rescheduled.
  * If 'time' is a null pointer, it will never yield.
*/
#define PT_YIELD_COOP(pt, time) \
do { \
    LC_SET((pt)->lc); \
    clock_gettime(CLOCK_REALTIME, &PT_YIELD_TIME); \
    if((time) && rtcp_timespec_less((time), &PT_YIELD_TIME)) \
        { return PT_YIELDED; } \
} while(0)

/** Loop while checking for a cooperative yield only every couple iterations.
  * Useful with thin loops where the overhead of a clock_gettime every iteration
  * would waste considerable performance compared to the loop body.
  *
  * Specify your loop body within a new block using curly brackets '{}' right
  * after this macro. Before the closing bracket, you must use PT_LOOP_INCR.
    \param start statement that initializes your loop
    \param end your loop condition expression
    \param counter a symbol to serve as an integer counter
    \param between how many increments should pass between yield checks
*/
#define PT_LOOP(start, end, counter, between) \
(start); \
while((end)) \
{ \
    (counter) = 0; \
    while((counter) < (between) && (end))

/** Increment loop while checking for a cooperative yield.
  * This macro is used in tandem with PT_LOOP.
    \param pt protothread
    \param time ending timepoint
    \param incr your loop increment statement
    \param counter a symbol to serve as an integer counter
*/
#define PT_LOOP_INCR(pt, time, incr, counter) \
        ++(counter); (incr); \
    } \
    PT_YIELD_COOP(pt, time);

#endif /* RTCP_H */
