#ifndef RTCP_SCHED_H
#define RTCP_SCHED_H

#define _POSIX_C_SOURCE 200112L
#include "rtcproto/posixver.h"

#include "rtcproto/proto.h"
#include "rtcproto/rtcproto.h"
#include "rtcproto/heap.h"
#include "rtcproto/hmap.h"
#include "rtcproto/uvec.h"

/** Type storing statistics associated with a certain protothread type */
typedef struct rtcp_stats_t
{
    /** Number of entries into / invocations of the protothead. */
    int entries;
    /** Total time spent in that protothread. */
    struct timespec total;
    /** Local continuation of the maximal overrun. */
    rtcp_lc_t max_lc;
    /** Time length of the maximal overrun. */
    struct timespec max_over;
} rtcp_stats_t;

void rtcp_stats_init(rtcp_stats_t* stats);
void rtcp_stats_update(rtcp_stats_t* stats, rtcp_lc_t lc, struct timespec actual, struct timespec overrun);

/** Scheduler call.
*/
typedef struct rtcp_call_t
{
    /** Expected invocation time (absolute timestamp). */
    struct timespec stamp;
    /** Sequence number. */
    unsigned int seq;
    /** Protothread to be invoked. */
    fn_prthread_t func;
    /** Protothread representation. */
    rtcp_pt_t pt;
    /** Index of the local state associated with the call in the local
        state storage associated with the function to be invoked. */
    int uvecidx;
} rtcp_call_t;

int rtcp_call_comp_less(void* a, void* b);

void rtcp_call_idx_update(void* val, int idx, void* user);

/** Prefix that serves as a header in memory for each local state. */
typedef struct rtcp_prefix_t
{
    /** Index of the protothread in the call queue the state belongs to. */
    int heapidx;
} rtcp_prefix_t;

void rtcp_prefix_idx_update(void* val, int idx, void* user);

/** Type storing local states associated with a certain protothread type. */
typedef struct rtcp_assoc_t
{
    /** Stores data blocks of (rtcp_prefix_t + local state). */
    rtcp_uvec_t locals;
    /** Run statistics. */
    rtcp_stats_t stats;
} rtcp_assoc_t;

void rtcp_assoc_init(rtcp_assoc_t* assoc);
void rtcp_assoc_malloc(rtcp_assoc_t* assoc, int typesize);
void rtcp_assoc_free(rtcp_assoc_t* assoc);
void rtcp_assoc_tick(rtcp_assoc_t* assoc);
void rtcp_assoc_shrink(rtcp_assoc_t* assoc);

/** Maximum number of scheduled calls that can be waiting at any one time. */
#define RTCP_SCHED_CALLS_MAX ((unsigned int)((((unsigned int)~0) >> 1) + 1))

/** Run-to-completion scheduler.
  * 
  * Uses sequence numbers to distinguish calls having equal timestamps, so they
  * can be invoked in a first-in first-out order. However, only up to half of
  * an unsigned int number of calls may be waiting at any one time, so that the
  * head and tail of the sequence can be ordered correctly.
  * 
  * Expands its memory pool as required, but shrinks it lazily.
*/
typedef struct rtcp_sched_t
{
    /** Call queue. Stores rtcp_call_t's. */
    rtcp_heap_t calls;
    /** Local state storages associated with protothread types.
        Stores rtcp_assoc_t's. */
    rtcp_hmap_t assoc;
    /** Sequence number to distinguish equal timestamps. */
    unsigned int seq;
} rtcp_sched_t;

int rtcp_sched_malloc(rtcp_sched_t* sched);
void rtcp_sched_free(rtcp_sched_t* sched);

void rtcp_sched_free_assoc(void* val);
void rtcp_sched_tick_assoc(void* val);
void rtcp_sched_shrink_assoc(void* val);

void rtcp_sched_tick(rtcp_sched_t* sched);
void rtcp_sched_shrink(rtcp_sched_t* sched);

int rtcp_sched_func_get_assoc(rtcp_sched_t* sched, fn_prthread_t func, int statesize, rtcp_assoc_t** out);
int rtcp_sched_calls_queued(rtcp_sched_t* sched);
int rtcp_sched_queue(rtcp_sched_t* sched, struct timespec stamp, fn_prthread_t func, void* state, int statesize);

/** Convenience macro, potential footgun.
  * Think of the state parameter as if you were passing it by value.
*/
#define rtcp_sched_mqueue(sched, stamp, func, state) rtcp_sched_queue(sched, stamp, func, &(state), sizeof(state))

void rtcp_sched_call(rtcp_sched_t* sched, struct timespec end, int endignore);
void rtcp_sched_until(rtcp_sched_t* sched, struct timespec end);
void rtcp_sched_flush(rtcp_sched_t* sched);
void rtcp_sched_flush_until(rtcp_sched_t* sched, struct timespec end);

void rtcp_sched_func_stats(rtcp_sched_t* sched, fn_prthread_t func, rtcp_stats_t* out);

#endif /* RTCP_SCHED_H */
