#include "rtcproto/heap.h"

void rtcp_heap_init(rtcp_heap_t* heap)
{
    /* Initialize values */
    heap->size = 0;
    heap->callback = NULL;
    heap->user = NULL;
}

void rtcp_heap_cb(rtcp_heap_t* heap, fn_idx_update_t callback, void* user)
{
    /* Initialize values */
    heap->callback = callback;
    heap->user = user;
}

int rtcp_heap_malloc(rtcp_heap_t* heap, int capacity, int typesize, fn_heap_compare_t compare)
{
    /* Store parameters */
    heap->capacity = capacity;
    rtcp_assert_always(heap->capacity > 0, "rtcp_heap_malloc: capacity > 0\n");
    heap->typesize = typesize;
    rtcp_assert_always(heap->typesize > 0, "rtcp_heap_malloc: typesize > 0\n");
    heap->compare = compare;
    rtcp_assert_always(heap->compare != NULL, "rtcp_heap_malloc: compare != NULL\n");

    /* Initialize values */
    heap->targetcapa = heap->capacity;

    /* Allocate memory */
    heap->values = malloc(heap->capacity * heap->typesize);
    if(!heap->values){ return errno; }
    heap->swap = malloc(heap->typesize);
    if(!heap->swap){ return errno; }

    return 0;
}

void rtcp_heap_free(rtcp_heap_t* heap)
{
    /* Free memory */
    assert(heap->values); free(heap->values);
    assert(heap->swap); free(heap->swap);
}

int rtcp_heap_realloc(rtcp_heap_t* heap, int newcapa)
{
    void* newvalues;

    /* Attempt to change the size of the allocated memory block */
    newvalues = realloc(heap->values, newcapa * heap->typesize);
    if(!newvalues){ return errno; }

    /* Attach block if allocation succeeded */
    heap->values = newvalues;

    /* Note new capacity */
    heap->capacity = newcapa;

    return 0;
}

void rtcp_heap_swap(rtcp_heap_t* heap, int a, int b)
{
    memcpy(
        heap->swap,
        ((uint8_t*)heap->values) + a * heap->typesize,
        heap->typesize
    );

    memcpy(
        ((uint8_t*)heap->values) + a * heap->typesize,
        ((uint8_t*)heap->values) + b * heap->typesize,
        heap->typesize
    );

    memcpy(
        ((uint8_t*)heap->values) + b * heap->typesize,
        heap->swap,
        heap->typesize
    );

    /* Value update callbacks */
    if(heap->callback){
        heap->callback(
            ((uint8_t*)heap->values) + a * heap->typesize,
            a,
            heap->user
        );
        heap->callback(
            ((uint8_t*)heap->values) + b * heap->typesize,
            b,
            heap->user
        );
    }
}

void rtcp_heap_tick(rtcp_heap_t* heap)
{
    /* Decrease target capacity until a lower bound */
    heap->targetcapa = heap->targetcapa > 1 ?
        heap->targetcapa - 1 : heap->targetcapa;
}

void rtcp_heap_use_size(rtcp_heap_t* heap)
{
    int cpot;

    assert(heap->targetcapa > 0);

    cpot = rtcp_ceil_power_of_two(
        heap->size > heap->targetcapa - 1 ? heap->size : heap->targetcapa - 1);

    /* Set target capacity to indicate usage */
    heap->targetcapa = cpot > 0 ? cpot : 1;
}

int rtcp_heap_shrink(rtcp_heap_t* heap)
{
    int ret;
    int cpot;

    cpot = rtcp_ceil_power_of_two(heap->size);

    assert(heap->targetcapa > 0);
    assert(heap->capacity > 0);

    /* If the target capacity less-or-equal than the next shrink */
    if(heap->targetcapa <= heap->capacity / 2){
        
        /* Reallocate with the log2-ceiling of the target */
        ret = rtcp_heap_realloc(heap, cpot > 0 ? cpot : 1);
        if(ret){ return ret; }
    }

    assert(heap->capacity > 0);
    assert(heap->capacity >= heap->size);

    return 0;
}

int rtcp_heap_property(rtcp_heap_t* heap)
{
    int i, j, j2;
    int prop;
    int desc;

    /* Heap property true for zero elements */
    prop = 1;

    /* Loop through first nodes of every level */
    for(i = 0; i < heap->size; i = i * 2 + 1){

        /* Loop through all nodes of the current level */
        j2 = i * 2 + 1; j2 = j2 < heap->size ? j2 : heap->size;
        for(j = i; j < j2; ++j){

            /* Compare root with descendants */
            desc = j * 2 + 1;
            if(desc < heap->size){
                prop = prop && heap->compare(
                    ((uint8_t*)heap->values) + j * heap->typesize,
                    ((uint8_t*)heap->values) + desc * heap->typesize
                );
            }
            desc = j * 2 + 2;
            if(desc < heap->size){
                prop = prop && heap->compare(
                    ((uint8_t*)heap->values) + j * heap->typesize,
                    ((uint8_t*)heap->values) + desc * heap->typesize
                );
            }
        }
    }

    return prop;
}

void rtcp_heap_swim(rtcp_heap_t* heap, int idx)
{
    int ancestor;

    /* Sift value upwards, until it either:
       - reaches the root node
       - does not violate the heap property relative to its ancestor
     */
    ancestor = (idx - 1) / 2;
    while(
        idx > 0 &&
        !heap->compare(
            ((uint8_t*)heap->values) + ancestor * heap->typesize,
            ((uint8_t*)heap->values) + idx * heap->typesize)
    ){
        rtcp_heap_swap(heap, ancestor, idx);

        idx = (idx - 1) / 2;
        ancestor = (idx - 1) / 2;
    }
}

void rtcp_heap_sink(rtcp_heap_t* heap, int idx)
{
    int desc1, desc2;

    /* Sift value upwards, until it either:
       - reaches the bottommost level
       - does not violate the heap property relative to its descendant
     */
    while(idx < heap->size){

        desc1 = idx * 2 + 1; desc2 = idx * 2 + 2;

        /* Select descendant based on comparison if there are two */
        if(
            desc2 < heap->size &&
            heap->compare(
                ((uint8_t*)heap->values) + desc2 * heap->typesize,
                ((uint8_t*)heap->values) + desc1 * heap->typesize)
        ){ desc1 = desc2; }

        /* If the selected descendant is valid, compare it, swap if necessary */
        if(
            desc1 < heap->size &&
            !heap->compare(
                ((uint8_t*)heap->values) + idx * heap->typesize,
                ((uint8_t*)heap->values) + desc1 * heap->typesize)
        ){
            rtcp_heap_swap(heap, idx, desc1);
            idx = desc1;
            continue;
        }

        /* If no swap occurred, the sink is complete */
        break;
    }
}

int rtcp_heap_insert(rtcp_heap_t* heap, void* val)
{
    int ret;

    /* If the new size exceeds the current capacity, reallocate */
    if(heap->size + 1 > heap->capacity){
        ret = rtcp_heap_realloc(heap, heap->capacity * 2);
        if(ret){ return ret; }
    }
    assert(heap->size + 1 <= heap->capacity);

    /* New size */
    ++heap->size;

    /* Write value */
    memcpy(
        ((uint8_t*)heap->values) + (heap->size - 1) * heap->typesize,
        val,
        heap->typesize
    );

    /* Value update callback */
    if(heap->callback){
        heap->callback(
            ((uint8_t*)heap->values) + (heap->size - 1) * heap->typesize,
            heap->size - 1,
            heap->user
        );
    }

    /* Sift up new element to restore the heap property */
    rtcp_heap_swim(heap, heap->size - 1);

    /* Use the current size, shrink if necessary */
    rtcp_heap_use_size(heap);
    rtcp_heap_shrink(heap);

    return 0;
}

void* rtcp_heap_peek(rtcp_heap_t* heap)
{
    return heap->size > 0 ? heap->values : NULL;
}

void rtcp_heap_extract(rtcp_heap_t* heap, void* val)
{
    rtcp_assert_always(heap->size > 0, "rtcp_heap_extract: empty heap\n");

    /* Peek only if the destination space is non-null */
    if(val){ memcpy(val, ((uint8_t*)heap->values), heap->typesize); }

    /* If there are remaining elements after the deletion */
    if(heap->size > 1){
        
        /* Place last element at the root */
        memcpy(
            ((uint8_t*)heap->values),
            ((uint8_t*)heap->values) + (heap->size - 1) * heap->typesize,
            heap->typesize
        );

        /* Value update callback */
        if(heap->callback){
            heap->callback(((uint8_t*)heap->values), 0, heap->user);
        }
    }

    /* New size */
    --heap->size;

    /* Sift down the element to restore the heap property */
    rtcp_heap_sink(heap, 0);

    /* Use the current size, shrink if necessary */
    rtcp_heap_use_size(heap);
    rtcp_heap_shrink(heap);
}

void* rtcp_heap_get(rtcp_heap_t* heap, int idx)
{
    rtcp_assert_always(idx >= 0 && idx < heap->size,
        "rtcp_heap_get: invalid index\n");

    return ((uint8_t*)heap->values) + idx * heap->typesize;
}
