#include "rtcproto/sched.h"

void rtcp_stats_init(rtcp_stats_t* stats)
{
    /* Initialize values */
    stats->entries = 0;
    stats->total.tv_sec = stats->total.tv_nsec = 0;
    stats->max_lc = 0;
    stats->max_over.tv_sec = stats->max_over.tv_nsec = 0;
}

void rtcp_stats_update(rtcp_stats_t* stats, rtcp_lc_t lc, struct timespec actual, struct timespec overrun)
{
    /* Update entry count */
    ++stats->entries;

    /* Update total runtime */
    stats->total = rtcp_timespec_add(&stats->total, &actual);

    /* Update maximum overrun */
    if(rtcp_timespec_less(&stats->max_over, &overrun)){
        stats->max_lc = lc;
        stats->max_over = overrun;
    }
}

int rtcp_call_comp_less(void* a, void* b)
{
    rtcp_call_t* ca = (rtcp_call_t*)a;
    rtcp_call_t* cb = (rtcp_call_t*)b;

    /* If the timestamps are equal, check the sequence numbers */
    if(
        ca->stamp.tv_sec == cb->stamp.tv_sec &&
        ca->stamp.tv_nsec == cb->stamp.tv_nsec
    ){
        /* Sequence numbers must be guaranteed to be nonidentical */
        assert(ca->seq != cb->seq);

        /* Wrapping comparison of sequence numbers:
           (a < b) should be true when: we try to increment each number
           to reach the other in a wrapping fashion, and we have to
           increment 'a' less than 'b' for that to happen. */
        return (unsigned int)(ca->seq - cb->seq) > RTCP_SCHED_CALLS_MAX;
    }
    /* Otherwise, just compare the timestamps themselves */
    else {
        return rtcp_timespec_less(
            &((rtcp_call_t*)a)->stamp,
            &((rtcp_call_t*)b)->stamp
        );
    }
}

void rtcp_call_idx_update(void* val, int idx, void* user)
{
    rtcp_call_t* call;
    rtcp_hmap_t* hmap;
    rtcp_prefix_t* prefix;
    rtcp_assoc_t* assoc;

    call = (rtcp_call_t*)val;
    hmap = (rtcp_hmap_t*)user;

    /* Get data associated with the call's function */
    assoc = rtcp_hmap_get(hmap, call->func);
    assert(assoc);

    /* Get data associated with the current call */
    prefix = (rtcp_prefix_t*)
        rtcp_uvec_get(&assoc->locals, call->uvecidx);
    
    /* Set the updated heap index */
    prefix->heapidx = idx;
}

void rtcp_prefix_idx_update(void* val, int idx, void* user)
{
    rtcp_prefix_t* prefix;
    rtcp_heap_t* heap;
    rtcp_call_t* call;

    prefix = (rtcp_prefix_t*)val;
    heap = (rtcp_heap_t*)user;

    /* Get the call associated with the current local state */
    call = rtcp_heap_get(heap, prefix->heapidx);
    assert(call);

    /* Set the updated vector index */
    call->uvecidx = idx;
}

void rtcp_assoc_init(rtcp_assoc_t* assoc)
{
    /* Initialize structure storing local states */
    rtcp_uvec_init(&assoc->locals);

    /* Initialize structure storing statistics */
    rtcp_stats_init(&assoc->stats);
}

void rtcp_assoc_malloc(rtcp_assoc_t* assoc, int typesize)
{
    /* Allocate local state */
    rtcp_uvec_malloc(&assoc->locals, 1, typesize);
}

void rtcp_assoc_free(rtcp_assoc_t* assoc)
{
    /* Free memory */
    rtcp_uvec_free(&assoc->locals);
}

void rtcp_assoc_tick(rtcp_assoc_t* assoc)
{
    /* Tick structure storing local states */
    rtcp_uvec_tick(&assoc->locals);
}

void rtcp_assoc_shrink(rtcp_assoc_t* assoc)
{
    /* Shrink structure storing local states */
    rtcp_uvec_shrink(&assoc->locals);
}

int rtcp_sched_malloc(rtcp_sched_t* sched)
{
    int ret;
    (void)sched;

    /* Allocate call queue */
    ret = rtcp_heap_malloc(&sched->calls,
        1, sizeof(rtcp_call_t), rtcp_call_comp_less);
    if(ret){ return ret; }
    rtcp_heap_init(&sched->calls);

    /* Allocate data associated to calls */
    ret = rtcp_hmap_malloc(&sched->assoc, 1, sizeof(rtcp_assoc_t));
    if(ret){ return ret; }
    rtcp_hmap_init(&sched->assoc);

    /* Initialize sequence number to distinguish equal-time timestamps */
    sched->seq = 0;

    return 0;
}

void rtcp_sched_free(rtcp_sched_t* sched)
{
    /* Free data structures associated with functions */
    rtcp_hmap_iter(&sched->assoc, rtcp_sched_free_assoc);

    /* Free memory */
    rtcp_heap_free(&sched->calls);
    rtcp_hmap_free(&sched->assoc);
}

void rtcp_sched_free_assoc(void* val)
{
    rtcp_assoc_free((rtcp_assoc_t*)val);
}

void rtcp_sched_tick_assoc(void* val)
{
    rtcp_assoc_tick((rtcp_assoc_t*)val);
}

void rtcp_sched_shrink_assoc(void* val)
{
    rtcp_assoc_shrink((rtcp_assoc_t*)val);
}

void rtcp_sched_tick(rtcp_sched_t* sched)
{
    /* Tick data structures associated with functions */
    rtcp_hmap_iter(&sched->assoc, rtcp_sched_tick_assoc);

    /* Tick immediate member structures */
    rtcp_heap_tick(&sched->calls);
    rtcp_hmap_tick(&sched->assoc);
}

void rtcp_sched_shrink(rtcp_sched_t* sched)
{
    /* Shrink data structures associated with functions */
    rtcp_hmap_iter(&sched->assoc, rtcp_sched_shrink_assoc);

    /* Shrink immediate member structures */
    rtcp_heap_shrink(&sched->calls);
    rtcp_hmap_shrink(&sched->assoc);
}

int rtcp_sched_func_get_assoc(rtcp_sched_t* sched, fn_prthread_t func, int statesize, rtcp_assoc_t** out)
{
    int ret;
    rtcp_assoc_t* assoc;
    rtcp_assoc_t newassoc;
    int localsize;

    rtcp_assert_always(statesize >= 0, "rtcp_sched_queue: statesize >= 0\n");

    /* Calculate actual size of local state */
    localsize = sizeof(rtcp_prefix_t) + statesize;

    /* Attempt to get the storage associated with the function */
    if(!(assoc = rtcp_hmap_get(&sched->assoc, func))){

        /* Allocate & initialize associated data structure */
        rtcp_assoc_malloc(&newassoc, localsize);
        rtcp_assoc_init(&newassoc);

        /* Add and retrieve associated data */
        ret = rtcp_hmap_add(&sched->assoc, func, &newassoc);
        if(ret){ return ret; }
        assoc = rtcp_hmap_get(&sched->assoc, func);
    }

    /* Assert that the size of the associated data is as expected */
    rtcp_assert_always(localsize == assoc->locals.typesize,
        "rtcp_sched_queue: non-matching statesize\n");

    /* Return associated data structure if it was requested */
    if(out){ *out = assoc; }

    return 0;
}

int rtcp_sched_calls_queued(rtcp_sched_t* sched)
{
    return sched->calls.size;
}

int rtcp_sched_queue(rtcp_sched_t* sched, struct timespec stamp, fn_prthread_t func, void* state, int statesize)
{
    int ret;
    rtcp_assoc_t* assoc;
    int uvecidx;
    rtcp_prefix_t* prefix;
    rtcp_call_t call;

    /* Get or create data structure associated with function */
    if((ret = rtcp_sched_func_get_assoc(sched, func, statesize, &assoc) )){
        return ret;
    }

    /* Get the index where the local state will be added to the vector */
    uvecidx = assoc->locals.size;

    /* Unset uvec callback */
    rtcp_uvec_cb(&assoc->locals, NULL, NULL);

    /* Add associated local state */
    ret = rtcp_uvec_add(&assoc->locals, NULL);
    if(ret){ return ret; }

    /* Write prefix portion of local state */
    prefix = (rtcp_prefix_t*)
        rtcp_uvec_get(&assoc->locals, assoc->locals.size - 1);
    assert(prefix);

    /* Write state portion of local state */
    memcpy(((uint8_t*)prefix) + sizeof(rtcp_prefix_t), state, statesize);

    /* Set heap callback */
    rtcp_heap_cb(&sched->calls, rtcp_call_idx_update, &sched->assoc);

    /* Schedule call */
    call.stamp = stamp;
    call.seq = sched->seq;
    call.func = func;
    PT_INIT(&call.pt);
    call.uvecidx = uvecidx;
    rtcp_heap_insert(&sched->calls, &call);

    /* Increment sequence number */
    ++sched->seq;

    return 0;
}

void rtcp_sched_call(rtcp_sched_t* sched, struct timespec end, int endignore)
{
    int status;
    rtcp_call_t* call;
    rtcp_prefix_t* prefix;
    rtcp_assoc_t* assoc;
    struct timespec limit;
    struct timespec run_start;
    struct timespec run_end;
    struct timespec time_zero = {0};

    /* Get next call in queue */
    call = rtcp_heap_peek(&sched->calls);

    /* Get data associated with the call's function */
    assoc = rtcp_hmap_get(&sched->assoc, call->func);
    assert(assoc);

    /* Get local state associated with call */
    prefix = (rtcp_prefix_t*)
        rtcp_uvec_get(&assoc->locals, call->uvecidx);
    assert(prefix->heapidx == 0);

    /* Copy time limit the protothread will receive a pointer to */
    limit = end;

    /* Execute call */
    clock_gettime(CLOCK_REALTIME, &run_start);
    status = call->func(
        &call->pt,
        ((uint8_t*)prefix) + sizeof(rtcp_prefix_t),
        endignore ? NULL : &limit
    );
    clock_gettime(CLOCK_REALTIME, &run_end);

    /* Assert that the return status is a recognizable one */
    rtcp_assert_always(status >= PT_WAITING && status <= PT_ENDED,
        "rtcp_sched_call: unrecognized return status from protothread\n");

    /* Remove call and its local state from queue if it has completed */
    if(status >= PT_EXITED){

        /* Set uvec callback */
        rtcp_uvec_cb(&assoc->locals, rtcp_prefix_idx_update, &sched->calls);

        /* Delete local state associated with call */
        rtcp_uvec_del(&assoc->locals, call->uvecidx);

        /* Extract call from queue */
        rtcp_heap_extract(&sched->calls, NULL);
    }

    /* Update statistics for the call's function */
    rtcp_stats_update(&assoc->stats,
        call->pt.lc,
        rtcp_timespec_diff(&run_start, &run_end),
        !endignore && rtcp_timespec_less(&end, &run_end) ?
            rtcp_timespec_diff(&end, &run_end) : time_zero
    );
}

void rtcp_sched_until(rtcp_sched_t* sched, struct timespec end)
{
    struct timespec now;

    /* Set heap callback */
    rtcp_heap_cb(&sched->calls, rtcp_call_idx_update, &sched->assoc);

    /* Schedule calls until either:
       - the designated timepoint has passed
       - there are no more calls waiting to run */
    while(
        rtcp_heap_peek(&sched->calls) &&
        clock_gettime(CLOCK_REALTIME, &now) == 0 &&
        rtcp_timespec_less(&now, &end)
    ){
        rtcp_sched_call(sched, end, 0);
    }
}

void rtcp_sched_flush(rtcp_sched_t* sched)
{
    struct timespec now;

    /* Set heap callback */
    rtcp_heap_cb(&sched->calls, rtcp_call_idx_update, &sched->assoc);

    /* Schedule calls until:
       - there are no more calls waiting to run */
    while(
        rtcp_heap_peek(&sched->calls)
    ){
        rtcp_sched_call(sched, now, 1);
    }
}

void rtcp_sched_flush_until(rtcp_sched_t* sched, struct timespec end)
{
    rtcp_call_t* call;

    /* Set heap callback */
    rtcp_heap_cb(&sched->calls, rtcp_call_idx_update, &sched->assoc);

    /* Schedule calls until either:
       - the next call is beyond the specified timepoint
       - there are no more calls waiting to run */
    while(
        (call = rtcp_heap_peek(&sched->calls)) &&
        rtcp_timespec_less(&call->stamp, &end)
    ){
        rtcp_sched_call(sched, end, 0);
    }
}

void rtcp_sched_func_stats(rtcp_sched_t* sched, fn_prthread_t func, rtcp_stats_t* out)
{
    rtcp_assoc_t* assoc;

    /* Get data associated with the protothread function */
    assoc = rtcp_hmap_get(&sched->assoc, func);
    rtcp_assert_always(assoc != NULL,
        "rtcp_sched_func_stats: unknown function\n");

    *out = assoc->stats;
}
