#include "../include/common.h"

int heap_int_less(void* a, void* b)
{
    return *(int*)a < *(int*)b;
}

void heap_callback_1(void* val, int idx, void* user)
{
    (void)val;
    (void)idx;

    /* Increment flag */
    ++(*(int*)user);
}

void heap_callback_2(void* val, int idx, void* user)
{
    int* expect;

    (void)val;
    
    /* Get expected index */
    expect = (int*)user;
    
    /* Check that the index moved into was the expected one */
    *expect = (*expect == idx);
}

TEST_IMPL(heap_insert)
{
    int ret;
    int val;

    rtcp_heap_t heap;
    ret = rtcp_heap_malloc(&heap, 1, sizeof(int), heap_int_less);
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_heap_init(&heap);

    val = 0x0BADF00D; ret = rtcp_heap_insert(&heap, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

    rtcp_heap_free(&heap);
    
    TEST_SUCCESS
}

TEST_IMPL(heap_insert_callback)
{
    int ret;
    int val;
	int flag = 0;

    rtcp_heap_t heap;
    ret = rtcp_heap_malloc(&heap, 1, sizeof(int), heap_int_less);
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_heap_init(&heap);
	rtcp_heap_cb(&heap, heap_callback_1, &flag);

    val = 0x0BADF00D; ret = rtcp_heap_insert(&heap, &val);
    ASSERT_EXT(ret == 0, "failed to add item");
    ASSERT_EXT(flag == 1, "callback counter expected 1");

    rtcp_heap_free(&heap);
    
    TEST_SUCCESS
}

TEST_IMPL(heap_property_insert)
{
    int ret;
    int val;
    int i;
    int n;

    rtcp_heap_t heap;
    ret = rtcp_heap_malloc(&heap, 1, sizeof(int), heap_int_less);
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_heap_init(&heap);
    
    n = 16;

    for(i = n; i > 0; --i){
		val = i; ret = rtcp_heap_insert(&heap, &val);
		ASSERT_EXT(ret == 0, "failed to add item");
		ASSERT_EXT(rtcp_heap_property(&heap), "insertion did not retain heap property");
    }

    rtcp_heap_free(&heap);
    
    TEST_SUCCESS
}

TEST_IMPL(heap_extract)
{
    int ret;
    int val;

    rtcp_heap_t heap;
    ret = rtcp_heap_malloc(&heap, 1, sizeof(int), heap_int_less);
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_heap_init(&heap);

    val = 0x0BADF00D; ret = rtcp_heap_insert(&heap, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

    val = 0x0D15EA5E; rtcp_heap_extract(&heap, &val);
    ASSERT_EXT(val == 0x0BADF00D, "wrong extract result");

    rtcp_heap_free(&heap);
    
    TEST_SUCCESS
}

TEST_IMPL(heap_extract_callback)
{
    int ret;
    int val;
	int expect;
	int flag;

    rtcp_heap_t heap;
    ret = rtcp_heap_malloc(&heap, 1, sizeof(int), heap_int_less);
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_heap_init(&heap);
	rtcp_heap_cb(&heap, heap_callback_2, &expect);

    val = 0x0BADF00D; ret = rtcp_heap_insert(&heap, &val);
    ASSERT_EXT(ret == 0, "failed to add item");
    val = 0x0BADC0DE; ret = rtcp_heap_insert(&heap, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

	expect = 0;
    val = 0x0D15EA5E; rtcp_heap_extract(&heap, &val);
    ASSERT_EXT(val == 0x0BADC0DE, "wrong extract result");
    ASSERT_EXT(expect != 0, "callback tester returned false (zero)");

	rtcp_heap_cb(&heap, heap_callback_1, &flag);
	flag = 0;
    val = 0x0D15EA5E; rtcp_heap_extract(&heap, &val);
    ASSERT_EXT(val == 0x0BADF00D, "wrong extract result");
    ASSERT_EXT(flag == 0, "callback should not have been called after last delete");

    rtcp_heap_free(&heap);
    
    TEST_SUCCESS
}

void heap_int_print(rtcp_heap_t* heap)
{
    int i, j;
    
    for(i = 0; i < heap->size; i = i * 2 + 1){
		for(j = i; j < i * 2 + 1 && j < heap->size; ++j){
			printf("%d ", ((int*)heap->values)[j]);
		}
		printf(" | ");
    }
    printf("\n");
}

TEST_IMPL(heap_property_extract)
{
    int ret;
    int val;
    int i;
    int n;

    rtcp_heap_t heap;
    ret = rtcp_heap_malloc(&heap, 1, sizeof(int), heap_int_less);
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_heap_init(&heap);
    
    n = 16;

    for(i = n; i > 0; --i){
		val = i; ret = rtcp_heap_insert(&heap, &val);
		ASSERT_EXT(ret == 0, "failed to add item");
		ASSERT_EXT(rtcp_heap_property(&heap), "insertion did not retain heap property");
    }

    for(i = 1; i < n + 1; ++i){
		rtcp_heap_extract(&heap, &val);
		ASSERT_EXT(val == i, "wrong extract result");
		ASSERT_EXT(rtcp_heap_property(&heap), "extraction did not retain heap property");
    }

    rtcp_heap_free(&heap);
    
    TEST_SUCCESS
}

TEST_IMPL(heap_shrink)
{
    int ret;
    int val;
    int i;
    int n, n2;

    rtcp_heap_t heap;
    ret = rtcp_heap_malloc(&heap, 1, sizeof(int), heap_int_less);
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_heap_init(&heap);
    
    /* Test with n items, where n is a power of two */
    n = n2 = 16;

    for(i = n; i > 0; --i){
		val = i; ret = rtcp_heap_insert(&heap, &val);
		ASSERT_EXT(ret == 0, "failed to add item");
		ASSERT_EXT(rtcp_heap_property(&heap), "insertion did not retain heap property");
    }

    for(i = 0; i < n; ++i){ rtcp_heap_tick(&heap); }

    rtcp_heap_extract(&heap, NULL); --n2;
    ASSERT_EXT(heap.capacity == n, "capacity should not have shrunk below used size");

    while(n2-- > n / 2){
		rtcp_heap_extract(&heap, &val);
		ASSERT_EXT(val == n - n2, "wrong extract result");
    }

    ASSERT_EXT(heap.capacity == n, "capacity should not have shrunk without ticking");
    ASSERT_EXT(heap.size == n / 2, "exactly half of the elements should have been removed");

    for(i = 0; i < n; ++i){ rtcp_heap_tick(&heap); }

    ASSERT_EXT(heap.capacity == n, "capacity should not have shrunk after ticking but no write op");

    for(i = 0; i < n; ++i){ rtcp_heap_tick(&heap); }

    rtcp_heap_extract(&heap, &val); --n2;
    ASSERT_EXT(heap.capacity == n / 2, "capacity should have shrunk after ticking and write op");

    while(n2-- >= 0){
		rtcp_heap_extract(&heap, &val);
		ASSERT_EXT(val == n - n2 - 1, "wrong extract result");
		for(i = 0; i < n; ++i){ rtcp_heap_tick(&heap); }
    }
    
    ASSERT_EXT(heap.capacity == 1, "capacity should have returned to 1 after removing all");

    rtcp_heap_free(&heap);
    
    TEST_SUCCESS
}

