#include "../include/common.h"

typedef struct StringState
{
	int n;
	char* str;
} StringState;

PT_THREAD(fn_proto(rtcp_pt_t* pt, void* state, struct timespec* end))
{
	char str[4];
	StringState* stt = (StringState*)state;
	(void)pt;
	(void)end;
	sprintf(str, "%d,", stt->n);
	strcat(stt->str, str);
	return PT_ENDED;
}

PT_THREAD(fn_proto_same(rtcp_pt_t* pt, void* state, struct timespec* end))
{
	char str[8];
	StringState* stt = (StringState*)state;
	(void)pt;
	(void)end;
	sprintf(str, "%d,", stt->n);
	str[0] = '\0';
	sprintf(str, "%d,", stt->n);
	strcat(stt->str, str);
	return PT_ENDED;
}

TEST_IMPL(sched_alloc)
{
	int ret;

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

TEST_IMPL(sched_queue)
{
	int ret;
	int state;
	struct timespec now;

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);

	ret = rtcp_sched_queue(&sched, now, fn_proto, &state, sizeof(int));
	ASSERT_EXT(ret == 0, "failed to allocate");

	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 1, "expected 1 queued call");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

TEST_IMPL(sched_until)
{
	int ret;
	char str[8*4] = {0};
	StringState state;
	struct timespec now;
	struct timespec then;
	int i, n;
	struct timespec offset = { 0, 10000000 };

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);

	state.str = str;
	n = 8;

	for(i = 0; i < n; ++i){
		state.n = i;
		ret = rtcp_sched_queue(&sched, now, fn_proto, &state, sizeof(StringState));
		ASSERT_EXT(ret == 0, "failed to allocate");
		rtcp_timespec_next(&now);
	}

	clock_gettime(CLOCK_REALTIME, &then);
	then = rtcp_timespec_add(&then, &offset);
	rtcp_sched_until(&sched, then);

	ASSERT_EXT(strcmp(str, "0,1,2,3,4,5,6,7,") == 0, "wrong result string");
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

TEST_IMPL(sched_until_reverse)
{
	int ret;
	char str[8*4] = {0};
	StringState state;
	struct timespec now;
	struct timespec then;
	int i, n;
	struct timespec offset = { 0, 10000000 };

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);

	state.str = str;
	n = 8;

	for(i = 0; i < n; ++i){
		state.n = i;
		ret = rtcp_sched_queue(&sched, now, fn_proto, &state, sizeof(StringState));
		ASSERT_EXT(ret == 0, "failed to allocate");
		rtcp_timespec_prev(&now);
	}

	clock_gettime(CLOCK_REALTIME, &then);
	then = rtcp_timespec_add(&then, &offset);
	rtcp_sched_until(&sched, then);

	ASSERT_EXT(strcmp(str, "7,6,5,4,3,2,1,0,") == 0, "wrong result string");
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

TEST_IMPL(sched_until_interleave)
{
	int ret;
	char str[8*4] = {0};
	StringState state;
	struct timespec now0;
	struct timespec now1;
	struct timespec now2;
	int i, n;

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now0);
	now1 = now0;
	now2 = now0; rtcp_timespec_next(&now2);

	state.str = str;
	n = 8;

	for(i = 0; i < n; i += 2){
		state.n = i;
		ret = rtcp_sched_queue(&sched, now1, fn_proto, &state, sizeof(StringState));
		ASSERT_EXT(ret == 0, "failed to allocate");
		rtcp_timespec_next(&now1);
		rtcp_timespec_next(&now1);
	}
	for(i = 1; i < n; i += 2){
		state.n = i;
		ret = rtcp_sched_queue(&sched, now2, fn_proto_same, &state, sizeof(StringState));
		ASSERT_EXT(ret == 0, "failed to allocate");
		rtcp_timespec_next(&now2);
		rtcp_timespec_next(&now2);
	}

	for(i = 0; i < n / 2; ++i){
		rtcp_timespec_next(&now0);
	}

	rtcp_sched_flush_until(&sched, now0);
	ASSERT_EXT(strcmp(str, "0,1,2,3,") == 0, "wrong result string halfway");
	rtcp_sched_flush(&sched);
	ASSERT_EXT(strcmp(str, "0,1,2,3,4,5,6,7,") == 0, "wrong result string");
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

TEST_IMPL(sched_flush)
{
	int ret;
	char str[8*4] = {0};
	StringState state;
	struct timespec now;
	int i, n;
	struct timespec offset = { 0, 10000000 };

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);
	now = rtcp_timespec_add(&now, &offset);

	state.str = str;
	n = 8;

	for(i = 0; i < n; ++i){
		state.n = i;
		ret = rtcp_sched_queue(&sched, now, fn_proto, &state, sizeof(StringState));
		ASSERT_EXT(ret == 0, "failed to allocate");
		rtcp_timespec_next(&now);
	}

	rtcp_sched_flush(&sched);

	ASSERT_EXT(strcmp(str, "0,1,2,3,4,5,6,7,") == 0, "wrong result string");
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

TEST_IMPL(sched_sequence)
{
	int ret;
	char str[8*4] = {0};
	StringState state;
	struct timespec now;
	int i, n;
	struct timespec offset = { 0, 10000000 };

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);
	now = rtcp_timespec_add(&now, &offset);

	state.str = str;
	n = 8;

	for(i = 0; i < n; ++i){
		state.n = i;
		ret = rtcp_sched_queue(&sched, now, fn_proto, &state, sizeof(StringState));
		ASSERT_EXT(ret == 0, "failed to allocate");
	}

	rtcp_sched_flush(&sched);

	ASSERT_EXT(strcmp(str, "0,1,2,3,4,5,6,7,") == 0, "wrong result string");
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

TEST_IMPL(sched_stats)
{
	int ret;
	char str[8*4] = {0};
	StringState state;
	struct timespec now;
	int i, n;
	struct timespec offset = { 0, 10000000 };
	rtcp_stats_t stats;

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	ret = rtcp_sched_func_get_assoc(&sched, fn_proto, sizeof(StringState), NULL);
	ASSERT_EXT(ret == 0, "failed to allocate");
	rtcp_sched_func_stats(&sched, fn_proto, &stats);
	ASSERT_EXT(stats.entries == 0, "incorrect entry count");

	clock_gettime(CLOCK_REALTIME, &now);
	now = rtcp_timespec_add(&now, &offset);

	state.str = str;
	n = 8;

	for(i = 0; i < n; ++i){
		state.n = i;
		ret = rtcp_sched_queue(&sched, now, fn_proto, &state, sizeof(StringState));
		ASSERT_EXT(ret == 0, "failed to allocate");
		rtcp_timespec_next(&now);
	}

	rtcp_sched_flush(&sched);

	rtcp_sched_func_stats(&sched, fn_proto, &stats);
	/*printf("entries: %d, total: %lu.%lu, max_lc: %u, max_over: %lu.%lu\n",
		stats.entries, stats.total.tv_sec, stats.total.tv_nsec,
		stats.max_lc, stats.max_over.tv_sec, stats.max_over.tv_nsec);*/
	ASSERT_EXT(stats.entries == n, "incorrect entry count");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}
